/*****************************************************************************\
* (c) Copyright 2018-2020 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the Apache License          *
* version 2 (Apache-2.0), copied verbatim in the file "LICENSE".              *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
#pragma once

#include "UTMagnetToolDefinitions.h"
#include "UTDefinitions.cuh"
#include "VeloDefinitions.cuh"
#include "CompassUTDefinitions.cuh"
#include "AlgorithmTypes.cuh"
#include "UTEventModel.cuh"

//=========================================================================
// Function definitions
//=========================================================================
namespace compass_ut {
  struct Parameters {
    HOST_INPUT(host_number_of_events_t, unsigned) host_number_of_events;
    DEVICE_INPUT(dev_number_of_events_t, unsigned) dev_number_of_events;
    DEVICE_INPUT(dev_ut_hits_t, char) dev_ut_hits; // actual hit contents
    DEVICE_INPUT(dev_ut_hit_offsets_t, unsigned) dev_ut_hit_offsets;
    DEVICE_INPUT(dev_velo_tracks_view_t, Allen::Views::Velo::Consolidated::Tracks) dev_velo_tracks_view;
    DEVICE_INPUT(dev_velo_states_view_t, Allen::Views::Physics::KalmanStates) dev_velo_states_view;
    DEVICE_INPUT(dev_ut_windows_layers_t, short) dev_ut_windows_layers;
    DEVICE_INPUT(dev_ut_number_of_selected_velo_tracks_with_windows_t, unsigned)
    dev_ut_number_of_selected_velo_tracks;
    DEVICE_INPUT(dev_ut_selected_velo_tracks_with_windows_t, unsigned) dev_ut_selected_velo_tracks;
    MASK_INPUT(dev_event_list_t) dev_event_list;
    DEVICE_OUTPUT(dev_ut_tracks_t, UT::TrackHits) dev_ut_tracks;
    DEVICE_OUTPUT(dev_offsets_ut_tracks_t, unsigned) dev_offsets_ut_tracks;
    HOST_OUTPUT(host_number_of_reconstructed_ut_tracks_t, unsigned) host_number_of_reconstructed_ut_tracks;

    PROPERTY(sigma_velo_slope_t, "sigma_velo_slope", "sigma velo slope [radians]", float) sigma_velo_slope;
    PROPERTY(min_momentum_final_t, "min_momentum_final", "final min momentum cut [MeV/c]", float) min_momentum_final;
    PROPERTY(min_pt_final_t, "min_pt_final", "final min pT cut [MeV/c]", float) min_pt_final;
    PROPERTY(hit_tol_2_t, "hit_tol_2", "hit_tol_2 [mm]", float) hit_tol_2;
    PROPERTY(delta_tx_2_t, "delta_tx_2", "delta_tx_2", float) delta_tx_2;
    PROPERTY(max_considered_before_found_t, "max_considered_before_found", "max_considered_before_found", unsigned)
    max_considered_before_found;
    PROPERTY(min_ld_3_hit_t, "min_ld_3_hit", "min_ld_3_hit", float) min_ld_3_hit;
    PROPERTY(min_ld_4_hit_t, "min_ld_4_hit", "min_ld_4_hit", float) min_ld_4_hit;
  };

  __global__ void compass_ut(
    Parameters,
    UTMagnetTool* dev_ut_magnet_tool,
    const float* dev_magnet_polarity,
    const float* dev_ut_dxDy,
    const unsigned* dev_unique_x_sector_layer_offsets);

  struct compass_ut_t : public DeviceAlgorithm, Parameters {
    void set_arguments_size(ArgumentReferences<Parameters> arguments, const RuntimeOptions&, const Constants&) const;

    void operator()(
      const ArgumentReferences<Parameters>& arguments,
      const RuntimeOptions&,
      const Constants& constants,
      const Allen::Context& context) const;

  private:
    Property<sigma_velo_slope_t> m_slope {this, 0.1f * Gaudi::Units::mrad};
    Property<min_momentum_final_t> m_mom_fin {this, 2500.f};
    Property<min_pt_final_t> m_pt_fin {this, 425.f};
    Property<hit_tol_2_t> m_hit_tol_2 {this, 0.8f * Gaudi::Units::mm};
    Property<delta_tx_2_t> m_delta_tx_2 {this, 0.018f};
    Property<max_considered_before_found_t> m_max_considered_before_found {this, 6};
    Property<min_ld_3_hit_t> m_min_ld_3_hit {this, -0.5f};
    Property<min_ld_4_hit_t> m_min_ld_4_hit {this, -0.5f};
  };
} // namespace compass_ut
