/*****************************************************************************\
* (c) Copyright 2018-2020 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the Apache License          *
* version 2 (Apache-2.0), copied verbatim in the file "LICENSE".              *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/

#include <MDFProvider.h>

MDFProvider::MDFProvider(
  size_t n_slices,
  size_t events_per_slice,
  std::optional<size_t> n_events,
  std::vector<std::string> connections,
  std::unordered_set<BankTypes> const& bank_types,
  MDFProviderConfig config = MDFProviderConfig {}) :
  InputProvider {n_slices, events_per_slice, bank_types, IInputProvider::Layout::Allen, n_events},
  m_buffer_status(n_slices), m_slice_to_buffer(n_slices, {-1, 0}), m_banks_version {n_slices},
  m_slice_free(n_slices, true), m_mfp_count {0}, m_event_ids {n_slices},
  m_connections {std::move(connections)}, m_config {config}
{
  // Preallocate prefetch buffer memory
  m_buffers.resize(n_slices);
  for (auto& [n_filled, event_offsets, buffer, transpose_start] : m_buffers) {
    auto epb = config.events_per_buffer;
    buffer.resize((epb < 100 ? 120 : epb + 20) * average_event_size * bank_size_fudge_factor * kB);
    event_offsets.resize(config.offsets_size + 20);
    event_offsets[0] = 0;
    n_filled = 0;
    transpose_start = 0;
  }

  // Initialize the current input filename
  m_current = m_connections.begin();

  // Allocate space to store event ids
  for (size_t n = 0; n < n_slices; ++n) {
    m_event_ids[n].reserve(events_per_slice + 20);
    // initialize bank version, needed for banks of subdetectors not present in input data
    std::fill(m_banks_version[n].begin(), m_banks_version[n].end(), -1);
  }

  m_odins.resize(n_slices);

  // Reserve 1MB for decompression
  m_compress_buffer.reserve(1u * MB);

  // Start prefetch thread and count bank types once a single buffer
  // is available
  {
    // aquire lock
    std::unique_lock<std::mutex> lock {m_prefetch_mut};

    m_masks.resize(n_slices);
    for (auto& mask : m_masks) {
      mask.resize(events_per_slice + 20, 0);
    }

    // start prefetch thread
    m_prefetch_thread = std::make_unique<std::thread>([this] { prefetch(); });

    // Wait for first read buffer to be full
    m_prefetch_cond.wait(lock, [this] { return !m_prefetched.empty() || m_read_error; });
    if (!m_read_error) {
      // Count number of banks per flavour
      bool count_success = false;

      // Offsets are to the start of the event, which includes the header
      auto i_read = m_prefetched.front();
      auto& [n_filled, event_offsets, buffer, transpose_start] = m_buffers[i_read];
      gsl::span<char> const event_span = {buffer.data(), event_offsets[1]};

      // Check what type of file we have: old MC or (new MC or data)
      m_is_mc = check_sourceIDs(event_span);
      if (*m_is_mc) {
        debug_cout << "Using bank types to determine subdetector\n";
        m_sd_from_raw = sd_from_bank_type;
        m_bank_sorter = sort_by_bank_type;
      }
      else {
        debug_cout << "Using 5 most significant bits to determine subdetector\n";
        m_sd_from_raw = sd_from_sourceID;
        m_bank_sorter = sort_by_sourceID;
      }
      std::tie(count_success, m_mfp_count) = fill_counts(event_span, m_sd_from_raw, m_config.skip_banks);

      if (!count_success) {
        error_cout << "Failed to determine bank counts\n";
        m_read_error = true;
      }
      else if (!m_read_error) {
        m_sizes_known = true;

        // Allocate slice memory that will contain transposed banks ready
        // for processing by the Allen kernels
        auto size_fun = [events_per_slice, this](BankTypes bank_type) -> std::tuple<size_t, size_t, size_t> {
          auto it = BankSizes.find(bank_type);
          auto ib = to_integral<BankTypes>(bank_type);
          if (it == end(BankSizes)) {
            throw std::out_of_range {std::string {"Bank type "} + std::to_string(ib) + " has no known size"};
          }
          // Allocate a minimum size
          auto allocate_events = events_per_slice < 100 ? 120 : events_per_slice + 20;

          // Allocate memory to store the offsets to the sizes and
          // the sizes themselves. The details of the memory layout
          // are documented in the Allen documentation.
          auto n_sizes = allocate_events * ((m_mfp_count[ib] + 2) / 2 + 1);

          // When events are transposed from the read buffer into
          // the per-rawbank-type slices, a check is made each time
          // to see if there is enough space available in a slice.
          // To avoid having to read every event twice to get the
          // size of all the banks.
          auto n_bytes = std::lround(
            ((1 + m_mfp_count[ib]) * sizeof(uint32_t) + it->second * kB) * allocate_events * bank_size_fudge_factor +
            2 * MB);
          return {n_bytes, n_sizes, allocate_events};
        };
        m_slices = allocate_slices(n_slices, types(), size_fun);
      }
    }
  }

  if (m_config.n_transpose_threads > n_slices) {
    debug_cout << "too many transpose threads requested with respect "
                  "to the number of read buffers; reducing the number of threads to "
               << n_slices;
    m_config.n_transpose_threads = n_slices;
  }

  // Start the transpose threads
  if (m_transpose_threads.empty() && !m_read_error) {
    for (size_t i = 0; i < m_config.n_transpose_threads; ++i) {
      m_transpose_threads.emplace_back([this, i] { transpose(i); });
    }
  }
}

MDFProvider::~MDFProvider()
{

  // Set flag to indicate the prefetch thread should exit, wake it
  // up and join it
  m_done = true;
  m_transpose_done = true;
  m_prefetch_cond.notify_one();
  if (m_prefetch_thread) m_prefetch_thread->join();

  // Set a flat to indicate all transpose threads should exit, wake
  // them up and join the threads. Ensure any waiting calls to
  // get_slice also return.
  m_prefetch_cond.notify_all();
  m_transpose_cond.notify_all();
  m_slice_cond.notify_all();

  for (auto& thread : m_transpose_threads) {
    thread.join();
  }
  release_buffers();
}

bool MDFProvider::release_buffers()
{
  free_slices(m_slices);
  return true;
}

EventIDs MDFProvider::event_ids(size_t slice_index, std::optional<size_t> first, std::optional<size_t> last) const
{
  auto const& ids = m_event_ids[slice_index];
  return {ids.begin() + (first ? *first : 0), ids.begin() + (last ? *last : ids.size())};
}

std::vector<char> MDFProvider::event_mask(size_t slice_index) const { return m_masks[slice_index]; }

BanksAndOffsets MDFProvider::banks(BankTypes bank_type, size_t slice_index) const
{
  auto ib = to_integral(bank_type);
  auto const& slice = m_slices[ib][slice_index];
  auto const& banks = slice.fragments;
  auto const offsets = slice.offsets;
  auto const offsets_size = slice.n_offsets;

  auto const version = m_banks_version[slice_index][ib];

  if (version == -1) {
    BanksAndOffsets bno {};
    bno.version = version;
    return bno;
  }
  else {
    gsl::span<char const> b {banks[0].data(), offsets[offsets_size - 1]};
    gsl::span<unsigned int const> s {slice.sizes.data(), slice.sizes.size()};
    gsl::span<unsigned int const> o {offsets.data(), static_cast<::offsets_size>(offsets_size)};
    gsl::span<unsigned int const> t {slice.types.data(), slice.types.size()};
    return BanksAndOffsets {
      {std::move(b)}, std::move(o), offsets[offsets_size - 1], std::move(s), std::move(t), version};
  }
}

std::tuple<bool, bool, bool, size_t, size_t, std::any> MDFProvider::get_slice(std::optional<unsigned int> timeout)
{
  bool timed_out = false, done = false;
  size_t slice_index = 0, n_filled = 0;
  std::any odin;
  std::unique_lock<std::mutex> lock {m_transpose_mut};
  if (!m_read_error) {
    // If no transposed slices are ready for processing, wait until
    // one is; use a timeout if requested
    if (m_transposed.empty()) {
      auto wakeup = [this] {
        auto n_writable = count_writable();
        return (!m_transposed.empty() || m_read_error || (m_transpose_done && n_writable == m_buffer_status.size()));
      };
      if (timeout) {
        timed_out = !m_transpose_cond.wait_for(lock, std::chrono::milliseconds {*timeout}, wakeup);
      }
      else {
        m_transpose_cond.wait(lock, wakeup);
      }
    }
    if (!m_read_error && !m_transposed.empty() && (!timeout || (timeout && !timed_out))) {
      std::tie(slice_index, n_filled) = m_transposed.front();
      m_transposed.pop_front();
      if (n_filled > 0) {
        odin = gsl::span<unsigned const> {m_odins[slice_index].data};
      }
    }
  }

  // Check if I/O and transposition is done and return a slice index
  auto n_writable = count_writable();
  done = m_transpose_done && m_transposed.empty() && n_writable == m_buffer_status.size();
  return {!m_read_error, done, timed_out, slice_index, n_filled, odin};
}

void MDFProvider::slice_free(size_t slice_index)
{
  // Check if a slice was acually in use before and if it was, only
  // notify the transpose threads that a free slice is available
  bool freed = false, set_writable = false;
  int i_read = 0;
  {
    std::unique_lock<std::mutex> lock {m_slice_mut};
    if (!m_slice_free[slice_index]) {
      m_slice_free[slice_index] = true;
      freed = true;

      // Reset the slice
      auto& event_ids = m_event_ids[slice_index];
      reset_slice(m_slices, slice_index, types(), event_ids);

      // Clear relation between slice and buffer
      i_read = m_slice_to_buffer[slice_index].buffer_index;
      auto& status = m_buffer_status[i_read];
      m_slice_to_buffer[slice_index].buffer_index = -1;

      if (
        status.work_counter == 0 && std::get<0>(m_buffers[i_read]) == std::get<3>(m_buffers[i_read]) &&
        (std::find_if(m_slice_to_buffer.begin(), m_slice_to_buffer.end(), [i_read](auto const stb) {
           return stb.buffer_index == i_read;
         }) == m_slice_to_buffer.end())) {
        status.writable = true;
        set_writable = true;
        // "Reset" buffer; the 0th offset is always 0. Set transpose
        // start to 0.
        std::get<0>(m_buffers[i_read]) = 0;
        std::get<3>(m_buffers[i_read]) = 0;
      }
    }
  }
  if (freed) {
    this->debug_output("Freed slice " + std::to_string(slice_index));
    m_odins[slice_index] = LHCb::ODIN {};
    m_slice_cond.notify_one();
  }
  if (set_writable) {
    this->debug_output("Set buffer " + std::to_string(i_read) + " writable");
    bool work_done = false;
    if (m_done && m_prefetched.empty()) {
      std::unique_lock<std::mutex> lock {m_prefetch_mut};
      work_done =
        std::accumulate(m_buffer_status.begin(), m_buffer_status.end(), 0u, [](size_t s, BufferStatus const& status) {
          return s + status.work_counter;
        }) == 0;
    }
    if (work_done && !m_transpose_done) {
      this->debug_output("Transpose done", 0);
      m_transpose_done = true;
      m_slice_cond.notify_all();
    }

    m_prefetch_cond.notify_one();
  }
}

gsl::span<char const>
MDFProvider::raw_banks(Allen::ReadBuffer const& buffer, size_t const read_event_start, size_t const event) const
{
  auto const event_index = event + read_event_start;

  // FIXME structured binding version below triggers clang 11 bug
  //       revert after clang fix available
  // auto const& [n_filled, event_offsets, event_buffer, transpose_start] = m_buffers[i_read];
  auto const& event_offsets = std::get<1>(buffer);
  auto const& event_buffer = std::get<2>(buffer);

  auto event_offset = event_offsets[event_index];

  // The first bank in the read buffer is the DAQ bank, which
  // contains the MDF header as bank payload; it does not belong to
  // the original event and should be skipped
  auto const* daq_bank = reinterpret_cast<LHCb::RawBank const*>(event_buffer.data() + event_offset);
  assert(daq_bank->type() == LHCb::RawBank::DAQ);
  auto const daq_bank_size = daq_bank->totalSize();

  auto const* banks_start = event_buffer.data() + event_offset + daq_bank_size;
  return {banks_start, event_offsets[event_index + 1] - event_offset - daq_bank_size};
}

void MDFProvider::event_sizes(
  size_t const slice_index,
  gsl::span<unsigned int const> const selected_events,
  std::vector<size_t>& sizes) const
{
  auto const stb = m_slice_to_buffer[slice_index];
  auto const i_read = stb.buffer_index;
  auto const read_event_start = stb.buffer_event_start;
  auto const& buffer = m_buffers[i_read];

  for (size_t i = 0; i < static_cast<size_t>(selected_events.size()); ++i) {
    auto const event = selected_events[i];
    sizes[i] = raw_banks(buffer, read_event_start, event).size_bytes();
  }
}

void MDFProvider::copy_banks(size_t const slice_index, unsigned int const event, gsl::span<char> output_buffer) const
{
  // The first bank in the read buffer is the DAQ bank, which
  // contains the MDF header as bank payload
  auto const stb = m_slice_to_buffer[slice_index];
  auto const i_read = stb.buffer_index;
  auto const read_event_start = stb.buffer_event_start;

  auto const& buffer = m_buffers[i_read];
  auto const banks = raw_banks(buffer, read_event_start, event);
  assert(banks.size_bytes() <= output_buffer.size());
  std::memcpy(output_buffer.data(), banks.data(), banks.size_bytes());
}

size_t MDFProvider::count_writable() const
{
  return std::accumulate(m_buffer_status.begin(), m_buffer_status.end(), 0ul, [](size_t s, BufferStatus const& stat) {
    return s + stat.writable;
  });
}

void MDFProvider::transpose(int thread_id)
{

  size_t i_read = 0;
  std::optional<size_t> slice_index {};

  bool good = false, transpose_full = false;
  size_t n_transposed = 0;

  while (!m_read_error && !m_transpose_done) {

    // Get a buffer to read from
    {
      std::unique_lock<std::mutex> lock {m_prefetch_mut};
      if (m_prefetched.empty() && !m_transpose_done) {
        m_prefetch_cond.wait(lock, [this] { return !m_prefetched.empty() || m_transpose_done; });
      }
      if (m_transpose_done) {
        break;
      }
      i_read = m_prefetched.front();
      m_prefetched.pop_front();
      this->debug_output("Got read buffer index " + std::to_string(i_read), thread_id);
      auto& status = m_buffer_status[i_read];
      status.writable = false;
      ++status.work_counter;
    }

    // Get a slice to write to
    if (!slice_index) {
      this->debug_output("Getting slice index", thread_id);
      auto it = m_slice_free.end();
      {
        std::unique_lock<std::mutex> lock {m_slice_mut};
        it = find(m_slice_free.begin(), m_slice_free.end(), true);
        if (it == m_slice_free.end()) {
          this->debug_output("Waiting for free slice", thread_id);
          m_slice_cond.wait(lock, [this, &it] {
            it = std::find(m_slice_free.begin(), m_slice_free.end(), true);
            return it != m_slice_free.end() || m_transpose_done;
          });
          // If transpose is done and there is no slice, we were
          // woken up be the desctructor before a slice was declared
          // free. In that case, exit without transposing
          if (m_transpose_done && it == m_slice_free.end()) {
            break;
          }
        }
        *it = false;
      }
      if (it != m_slice_free.end()) {
        slice_index = distance(m_slice_free.begin(), it);
        this->debug_output("Got slice index " + std::to_string(*slice_index), thread_id);
      }
    }

    // Transpose the events in the read buffer into the slice
    std::tie(good, transpose_full, n_transposed) = transpose_events(
      m_buffers[i_read],
      m_slices,
      *slice_index,
      this->types(),
      m_sd_from_raw,
      m_bank_sorter,
      m_mfp_count,
      m_config.skip_banks,
      m_banks_version[*slice_index],
      m_event_ids[*slice_index],
      m_masks[*slice_index],
      m_config.split_by_run);

    if (m_read_error || !good) {
      m_read_error = true;
      auto& status = m_buffer_status[i_read];
      --status.work_counter;
      m_transpose_cond.notify_one();
      break;
    }
    else {
      m_slice_to_buffer[*slice_index] = {static_cast<int>(i_read), std::get<3>(m_buffers[i_read])};
      auto ib = to_integral(BankTypes::ODIN);
      auto& slice = m_slices[ib][*slice_index];
      auto ob = odin_bank<false>(slice.fragments[0].data(), slice.offsets.data(), slice.sizes.data(), 0);
      m_odins[*slice_index] = MDF::decode_odin({ob.data, ob.size}, m_banks_version[*slice_index][ib]);
    }

    // Increment the transpose_start with the number of transposed events
    std::get<3>(m_buffers[i_read]) += n_transposed;
    auto events_left = std::get<0>(m_buffers[i_read]) - std::get<3>(m_buffers[i_read]);
    this->debug_output(
      "Transposed: slice_index " + std::to_string(*slice_index) + "; full " + std::to_string(transpose_full) +
        "; n_transposed " + std::to_string(n_transposed) + "; n_left " + std::to_string(events_left),
      thread_id);

    // Notify any threads waiting in get_slice that a slice is available
    {
      std::unique_lock<std::mutex> lock {m_transpose_mut};
      m_transposed.emplace_back(*slice_index, n_transposed);
    }
    m_transpose_cond.notify_one();

    slice_index.reset();

    // Check if the read buffer is now empty. If it is, it can be
    // reused, otherwise give it to another transpose thread once a
    // new target slice is available
    {
      std::unique_lock<std::mutex> lock {m_prefetch_mut};
      if (events_left > 0) {
        // Put this prefetched slice back on the prefetched queue so
        // somebody else can finish it
        m_prefetched.push_front(i_read);
      }

      // Decrement the work counter of this buffer
      auto& status = m_buffer_status[i_read];
      --status.work_counter;
    }
  }
}

bool MDFProvider::open_file() const
{
  bool good = false;

  // Check if there are still files available
  while (!good) {
    // If looping on input is configured, do it
    if (m_current == m_connections.end()) {
      if (++m_loop < m_config.n_loops) {
        m_current = m_connections.begin();
      }
      else {
        break;
      }
    }

    if (m_input && m_input->good) m_input->close();

    m_input = MDF::open(m_current->c_str(), O_RDONLY);
    if (m_input->good) {
      // read the first header, needed by subsequent calls to read_events
      ssize_t n_bytes = m_input->read(reinterpret_cast<char*>(&m_header), mdf_header_size);
      good = (n_bytes > 0);
    }

    if (good) {
      info_cout << "Opened " << *m_current << "\n";
    }
    else {
      error_cout << "Failed to open " << *m_current << " " << strerror(errno) << "\n";
      m_read_error = true;
      return false;
    }
    ++m_current;
  }
  return good;
}

void MDFProvider::prefetch()
{

  bool eof = false, error = false, prefetch_done = false;
  size_t bytes_read = 0;

  auto to_read = this->n_events();
  size_t eps = this->events_per_slice();

  // Loop while there are no errors and the flag to exit is not set
  while (!m_done && !m_read_error && (!to_read || *to_read > 0)) {

    // open the first file
    if (!m_input && !open_file()) {
      m_read_error = true;
      m_prefetch_cond.notify_one();
      return;
    }

    // Obtain a prefetch buffer to read into, if none is available,
    // wait until one of the transpose threads is done with its
    // prefetch buffer
    auto it = m_buffer_status.end();
    {
      auto find_writable = [this] {
        return std::find_if(
          m_buffer_status.begin(), m_buffer_status.end(), [](const auto& stat) { return stat.writable; });
      };

      std::unique_lock<std::mutex> lock {m_prefetch_mut};
      it = find_writable();
      if (it == m_buffer_status.end()) {
        this->debug_output("Waiting for free buffer");
        m_prefetch_cond.wait(
          lock, [this, &find_writable] { return (find_writable() != m_buffer_status.end()) || m_done; });
        if (m_done) {
          break;
        }
        else {
          it = find_writable();
        }
      }
      // Flag the prefetch buffer as unavailable
      it->writable = false;
    }
    size_t i_buffer = distance(m_buffer_status.begin(), it);
    auto& read_buffer = m_buffers[i_buffer];

    // Read events into the prefetch buffer, open new files as
    // needed
    while (true) {
      size_t read = std::get<0>(read_buffer);
      size_t to_prefetch = to_read ? std::min(eps, *to_read + read) : eps;
      std::tie(eof, error, bytes_read) =
        read_events(*m_input, read_buffer, m_header, m_compress_buffer, to_prefetch, m_config.check_checksum);

      size_t n_read = std::get<0>(read_buffer) - read;
      if (to_read) {
        *to_read -= std::min(*to_read, n_read);
      }

      if (error) {
        // Error encountered
        m_read_error = true;
        break;
      }
      else if (to_read && *to_read == 0) {
        if (m_config.n_loops != 0 && m_loop < (m_config.n_loops - 1)) {
          // Set things such that the next call to open_file will
          // result in a loop
          this->debug_output("Loop " + std::to_string(m_loop + 1));
          to_read = this->n_events();
          m_current = m_connections.end();
          if (m_input && m_input->good) m_input->close();
          m_input.reset();
        }
        else {
          // No events left to read
          this->debug_output("Prefetch done: n_events reached");
          prefetch_done = true;
        }
        break;
      }
      else if (std::get<0>(read_buffer) >= eps) {
        // Number of events in a slice reached
        break;
      }
      else if (eof && !open_file()) {
        // Try to open the next file, if there is none, prefetching
        // is done.
        if (!m_read_error) {
          this->debug_output("Prefetch done: eof and no more files");
        }
        prefetch_done = true;
        break;
      }
    }

    auto const n_events = std::get<0>(read_buffer);
    this->debug_output("Read " + std::to_string(n_events) + " events into " + std::to_string(i_buffer));

    if (m_is_mc && n_events > 0) {
      auto const is_mc = check_sourceIDs({std::get<2>(read_buffer).data(), std::get<1>(read_buffer)[1]});
      if (*m_is_mc != is_mc) {
        throw std::out_of_range {"The next batch of events is different from the previous events "s +
                                 (*m_is_mc ? "some banks now"s : "none of the banks"s) + "have the top 5 bits set"s};
      }
    }

    // Notify a transpose thread that a new buffer of events is
    // ready. If prefetching is done, wake up all threads
    if (!error) {
      {
        std::unique_lock<std::mutex> lock {m_prefetch_mut};
        if (n_events > 0) {
          m_prefetched.push_back(i_buffer);
        }
        else {
          it->writable = true;
        }
      }
      if (prefetch_done) {
        m_done = prefetch_done;
        this->debug_output("Prefetch notifying all");
        m_prefetch_cond.notify_all();
      }
      else {
        this->debug_output("Prefetch notifying one");
        m_prefetch_cond.notify_one();
      }
    }
  }
  m_prefetch_cond.notify_one();
}
